from flask.signals import Namespace as SignalNamespace


_signals = SignalNamespace()
model_payment_method_updated = _signals.signal('stripe_model_payment_method_updated')
model_subscription_updated = _signals.signal('stripe_model_subscription_updated')
model_subscription_invoice_created = _signals.signal('stripe_model_subscription_invoice_created')
model_last_invoice_updated = _signals.signal('stripe_model_last_invoice_updated')
invoice_payment = _signals.signal('stripe_invoice_payment')


def stripe_event_signal(event_name):
    return _signals.signal('stripe_%s' % event_name)


def connect_to_stripe_event(event_name):
    return stripe_event_signal(event_name).connect



from . import enqueue_now, enqueue_task, enqueue_task_now, get_enqueued_job, logger
from .job import FrascoJob
from rq import get_current_job
from frasco.ext import get_extension_state
from contextlib import contextmanager
import uuid


def connect_as_background_task(signal, **kwargs):
    def decorator(func):
        kwargs['weak'] = False
        def listener(sender, **kwargs):
            enqueue_task(func, sender, **kwargs)
        signal.connect(listener, **kwargs)
        return func
    return decorator


def _parallel_tasks_callback(identifier, job_id, callback, args, kwargs):
    redis = get_extension_state('frasco_tasks').rq.connection
    remaining_jobs = redis.scard(identifier)
    remaining_jobs -= redis.srem(identifier, job_id)
    if remaining_jobs > 0:
        return
    redis.delete(identifier)
    callback(*args, **kwargs)


@contextmanager
def parallel_tasks_callback(callback, *callback_args, identifier=None, **callback_kwargs):
    redis = get_extension_state('frasco_tasks').rq.connection
    if not identifier:
        identifier = str(uuid.uuid4())

    def enqueue(*args, **kwargs):
        job = enqueue_task_now(*args, **kwargs)
        redis.sadd(identifier, job.id)
        return enqueue_now(_parallel_tasks_callback, args=(identifier, job.id, callback, callback_args, callback_kwargs), depends_on=job)

    yield enqueue


def cancel_parallel_tasks(identifier, delete=True):
    redis = get_extension_state('frasco_tasks').rq.connection
    job_ids = redis.smembers(identifier)
    if not job_ids:
        return
    jobs = FrascoJob.fetch_many([i.decode('utf-8') for i in job_ids], connection=redis)
    logger.debug("Cancelling %s parallel tasks identified by %s" % (len(jobs), identifier))
    for job in jobs:
        if not job:
            continue
        if delete:
            job.delete(delete_dependents=True)
        elif not job.ended_at:
            job.cancel()
    redis.delete(identifier)
    return jobs



from flask_sqlalchemy import SQLAlchemy, Model as BaseModel
from flask_migrate import Migrate
from frasco.ext import get_extension_config
from frasco.helpers import inject_app_config


class Model(BaseModel):
    def __taskdump__(self):
        return None, str(self.id)

    @classmethod
    def __taskload__(cls, id):
        return cls.query.get(id)


class FrascoModels(SQLAlchemy):
    name = "frasco_models"

    def __init__(self, *args, **kwargs):
        kwargs['model_class'] = Model
        kwargs.setdefault('session_options', {}).setdefault('expire_on_commit', False)
        super(FrascoModels, self).__init__(*args, **kwargs)

    def init_app(self, app):
        app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
        inject_app_config(app, get_extension_config(app, 'frasco_models'), prefix="SQLALCHEMY_")
        super(FrascoModels, self).init_app(app)
        self.migrate = Migrate(app, self)


db = FrascoModels()




from flask import current_app
from frasco.templating import jinja_fragment_extension
from frasco.ext import get_extension_state


@jinja_fragment_extension("cache")
def CacheFragmentExtension(caller=None, key=None, timeout=None):
    state = get_extension_state('frasco_redis')
    rv = state.connection.get(key)
    if rv is None:
        timeout = timeout or state.options["fragment_cache_timeout"]
        rv = caller()
        state.connection.setex(key, timeout, rv)
    return rv
